unit Main;

{$WARN UNIT_PLATFORM OFF}

interface

uses
  Kernel, Platforms, Objects, Images, Windows, SysUtils, Variants, Classes,
  Graphics, Controls, Forms, Dialogs, About,
  XPMan, StdCtrls, Buttons, ShellCtrls, ExtCtrls, ComCtrls;

type
  TMainForm = class(TForm)
    XPMan: TXPManifest;
    GroupBox1: TGroupBox;
    GroupBox2: TGroupBox;
    GroupBox3: TGroupBox;
    StartBtn: TBitBtn;
    PauseBtn: TBitBtn;
    StopBtn: TBitBtn;
    BitBtn4: TBitBtn;
    InputEdit: TEdit;
    ShellTree: TShellTreeView;
    DirEdit: TLabeledEdit;
    ImageEdit: TLabeledEdit;
    Label1: TLabel;
    OpenD: TOpenDialog;
    GroupBox4: TGroupBox;
    BufferEdit: TLabeledEdit;
    PriorityCombo: TComboBox;
    Label2: TLabel;
    Label3: TLabel;
    FileEd: TEdit;
    Label4: TLabel;
    SizeEd: TEdit;
    Label5: TLabel;
    Label6: TLabel;
    FPosEd: TEdit;
    BPosEd: TEdit;
    Label7: TLabel;
    JPEGCountEd: TEdit;
    Label8: TLabel;
    LastEd: TEdit;
    Label9: TLabel;
    StatusEd: TEdit;
    Label10: TLabel;
    IFFCombo: TComboBox;
    Label11: TLabel;
    BitBtn1: TBitBtn;
    procedure Button1Click(Sender: TObject);
    procedure StartBtnClick(Sender: TObject);
    procedure PauseBtnClick(Sender: TObject);
    procedure StopBtnClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure BitBtn4Click(Sender: TObject);
  public
    InputFile: string;
    ParentOutput: string;
    DirOutput: string;
    Output: string;
    JPEGName: string;
    BufSize: Long;
    Completed: Boolean;
    Priority: Long;
    ImgFormat: Long;
  end;

  TProcessThread = class(TThread)
  protected
    FFileName: string;
    FFilePos: Long;
    FBuffer: Pointer;
    FBufSize: Long;
    FMaxBufSize: Long;
    FBufPos: Long;
    FStatus: string;
    FJPEGInfo: string;
    FCount: Long;
    FImg: TImageData;
    FMod: TNImageFileFormat;

    procedure UpdatePos;
    procedure UpdateStatus;
    procedure UpdateJPEGs;
    procedure Complete;
    procedure StopMe;
  public
    Stream: TFileStream;
    Pattern: string;
    Output: string;
    DefaultName: string;
    IFFExt: string;

    constructor Create(const AFileName: string; ABufSize: Long);
    destructor Destroy; override;
    procedure FreeResources;
    procedure Execute; override;
  end;

type
  TBMTable = array[0..255] of Long;

const
  JPEGPattern = #$FF#$D8#$FF;

var
  MainForm: TMainForm;
  Process: TProcessThread;


implementation

{$R *.dfm}


{*************************************************************************}
{                           Misc functions                                }
{*************************************************************************}

{Boyer-Moore string functions are copyrighted by Peter Morris.}

procedure MakeTable(Buffer: PChar; Len: Long; var Table: TBMTable);
begin
  if Len = 0 then Exit;
  asm
        push    EDI
        push    ESI

        mov     EDI, Table
        mov     EAX, Len
        mov     ECX, $100
        repne   STOSD

        mov     ECX, Len
        mov     EDI, Table
        mov     ESI, Buffer
        dec     ECX
        xor     EAX, EAX
@@loop:
        mov     AL, [ESI]
        lea     ESI, ESI + 1
        mov     [EDI + EAX * 4], ECX
        dec     ECX
        jg      @@loop

        pop     ESI
        pop     EDI
  end;
end;

function BMPos(const Source, Find: Pointer; const SourceLen, FindLen: Long; var
  Table: TBMTable): Pointer;
var
  LastPos: Pointer;
begin
  LastPos := Pointer(Long(Source) + SourceLen - 1);
  asm
        push    ESI
        push    EDI
        push    EBX

        mov     EAX, FindLen
        mov     ESI, Source
        lea     ESI, ESI + EAX - 1
        std
        mov     EBX, Table

@@comparetext:
        cmp     ESI, LastPos
        jg      @@NotFound
        mov     EAX, FindLen
        mov     EDI, Find
        mov     ECX, EAX
        push    ESI
        lea     EDI, EDI + EAX - 1
        xor     EAX, EAX
@@CompareNext:
        mov     al, [ESI]
        cmp     al, [EDI]
        jne     @@LookAhead
        lea     ESI, ESI - 1
        lea     EDI, EDI - 1
        dec     ECX
        jz      @@Found
        jmp     @@CompareNext

@@LookAhead:
        pop     ESI
        mov     al, [ESI]
        mov     EAX, [EBX + EAX * 4]
        lea     ESI, ESI + EAX
        jmp     @@CompareText

@@NotFound:
        mov     Result, 0
        jmp     @@TheEnd
@@Found:
        pop     EDI
        inc     ESI
        mov     Result, ESI
@@TheEnd:
        cld
        pop     EBX
        pop     EDI
        pop     ESI
  end;
end;

function FastPos(const Source, Find: string; const SourceLen, FindLen,
  StartPos: Long): Long;
var
  Table: TBMTable;
begin
  MakeTable(PChar(Find), FindLen, Table);
  Result := Long(BMPos(PChar(Long(Source) + (StartPos - 1)), PChar(Find),
    SourceLen - (StartPos - 1), FindLen, Table));
  if Result > 0 then
    Result := Result - Long(@Source[1]) + 1;
end;


{*************************************************************************}
{                          TMainForm methods                              }
{*************************************************************************}


procedure TMainForm.Button1Click(Sender: TObject);
begin
  if OpenD.Execute then
  begin
    InputFile := OpenD.FileName;
    InputEdit.Text := InputFile;
    InputEdit.SelStart := Length(InputFile);
  end;
end;

procedure TMainForm.StartBtnClick(Sender: TObject);
begin
  if Process <> nil then
    if Process.Suspended and (not Process.Terminated) then
    begin
      PauseBtn.Enabled := True;
      StopBtn.Enabled := True;
      Process.Resume;
      Exit;
    end;
  if not FileExists(InputFile) then
  begin
    Platform.MsgBox('Invalid input file: "' + InputFile + '"');
    Exit;
  end;
  if SameStr('.mpq', ExtractFileExt(InputFile)) then
  begin
    Platform.MsgBox('You are trying to open Blizzard''s ' +
      'MPQ archive (Warcraft 3?). Unfortunately, JPG files ' +
      'in these archives are not in standard format or ' +
      'they are encrypted somehow. I don''t know how to extract them ' +
      'correctly. Sorry.');
    Exit;
  end;
  ParentOutput := ShellTree.Path;
  if not DirectoryExists(ParentOutput) then
  begin
    Platform.MsgBox('Invalid output directory: "' + ParentOutput + '"');
    Exit;
  end;
  DirOutput := DirEdit.Text;
  Output := ParentOutput + PathDelim + DirOutput;
  if not ForceDirectories(Output) then
  begin
    Platform.MsgBox('Error when creating output directory: "' + Output + '"');
    Exit;
  end;
  try
    BufSize := StrToInt(BufferEdit.Text);
  except
    BufSize := 1024;
  end;
  Priority := PriorityCombo.ItemIndex;
  ImgFormat := IFFCombo.ItemIndex;

  FreeAndNil(Process);
  Process := TProcessThread.Create(InputFile, BufSize * 1024);
  Process.Pattern := JPEGPattern;
  Process.Output := Output;
  Process.DefaultName := ImageEdit.Text;

  case Priority of
    0: Process.Priority := tpLower;
    1: Process.Priority := tpNormal;
    2: Process.Priority := tpHigher;
    3: Process.Priority := tpTimeCritical;
  end;

  case ImgFormat of
    0: Process.IFFExt := 'jpg';
    1: Process.IFFExt := 'png';
    2: Process.IFFExt := 'tga';
    3: Process.IFFExt := 'bmp';
  end;

  FileEd.Text := ExtractFilename(InputFile);
  SizeEd.Text := IntToStr(Process.Stream.Size div 1024) + ' KB';
  JPEGCountEd.Text := '0';
  LastEd.Text := 'None';

  PauseBtn.Enabled := True;
  StopBtn.Enabled := True;
  Log.Add('JPEG extraction started from "' + InputFile + '"');
  Process.Resume;
end;

procedure TMainForm.PauseBtnClick(Sender: TObject);
begin
  if Process <> nil then
  begin
    MainForm.StatusEd.Text := 'Process paused! Press Start to continue.';
    Process.Suspend;
    PauseBtn.Enabled := False;
    StopBtn.Enabled := False;
  end;
end;

procedure TMainForm.StopBtnClick(Sender: TObject);
begin
  if Process = nil then Exit;
  if not Completed then
    MainForm.StatusEd.Text := 'Process stopped incompleted!';
  Completed := False;
  Process.Terminate;
  PauseBtn.Enabled := False;
  StopBtn.Enabled := False;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  StopBtnClick(Self);
  FreeAndNil(Process);
end;

procedure TMainForm.BitBtn4Click(Sender: TObject);
begin
  AboutForm.ShowModal;
end;


{*************************************************************************}
{                      TProcessThread methods                             }
{*************************************************************************}


procedure TProcessThread.Complete;
begin
  MainForm.Completed := True;
end;

constructor TProcessThread.Create(const AFileName: string; ABufSize: Long);
begin
  inherited Create(True);
  FFileName := AFileName;
  Stream := TFileStream.Create(FFileName, fmOpenRead);
  FMaxBufSize := ABufSize;
  GetMem(FBuffer, FMaxBufSize);
  FBufSize := Stream.Read(FBuffer^, FMaxBufSize);
  FMod := Images.FindImageFileFormat('jpg');
end;

destructor TProcessThread.Destroy;
begin
  FreeResources;
  inherited Destroy;
end;

procedure TProcessThread.Execute;
var
  OldPos: Long;
  S: string;
begin
  inherited;
  while not Terminated do
  begin
    FStatus := 'Searching for pattern';
    Synchronize(UpdateStatus);

    FBufPos := FastPos(string(FBuffer), Pattern + #$E0,
      FBufSize, Length(Pattern) + 1, FBufPos + 1);

    if FBufPos = 0 then
      FBufPos := FastPos(string(FBuffer), Pattern + #$E1,
        FBufSize, Length(Pattern) + 1, FBufPos + 1);

    if FBufPos = 0 then
    begin
      FStatus := 'Refilling buffer';
      Synchronize(UpdateStatus);
      if Stream.Position < Stream.Size - 1 then
        FBufSize := Stream.Read(FBuffer^, FMaxBufSize)
      else
      begin
        Synchronize(Complete);
        FStatus := 'Process complete!';
        Synchronize(UpdateStatus);
        Log.Add('JPEG extraction completed from "'
          + FFileName + '" JPEGs found: ' + IntToStr(FCount));
        Synchronize(StopMe);
        Continue;
      end;
    end
    else
    begin
      Stream.Position := Stream.Position - FBufSize + FBufPos - 1;
      OldPos := Stream.Position + Length(Pattern);
      FStatus := 'Decoding JPEG image';
      Synchronize(UpdateStatus);
      FreeImage(FImg);

      FMod.LoadImage(FImg, Stream);

      S := IntToStr(FCount);
      if Length(S) < 5 then
        S := StringOfChar('0', 5 - Length(S)) + S;

      Images.FindImageFileFormat(IFFExt).SaveImage(FImg,
        Output + PathDelim + DefaultName + S + '.' + IFFExt);

      FCount := FCount + 1;
      if FImg.Count > 0 then
        with FImg.Levels[0] do
          FJPEGInfo := Format('%dx%d %dbpp', [Width, Height,
            FImg.DataFormat.BitCount])
      else
        FJPEGInfo := 'Invalid';
      Synchronize(UpdateJPEGs);

      Log.Add('JPEG saved: ' + FJPEGInfo);

      Stream.Position := OldPos;
      FBufSize := Stream.Read(FBuffer^, FMaxBufSize);
      FBufPos := 0;
    end;
    Synchronize(UpdatePos);
  end;
end;

procedure TProcessThread.FreeResources;
begin
  FreeMemNil(FBuffer);
  FreeAndNil(Stream);
end;

procedure TProcessThread.StopMe;
begin
  MainForm.StopBtnClick(nil);
end;

procedure TProcessThread.UpdateJPEGs;
begin
  MainForm.JPEGCountEd.Text := IntToStr(FCount);
  MainForm.LastEd.Text := FJPEGInfo;
end;

procedure TProcessThread.UpdatePos;
begin
  MainForm.FPosEd.Text := IntToStr(Stream.Position div 1024) + ' KB';
  MainForm.BPosEd.Text := IntToStr(FBufPos) + '/' + IntToStr(FBufSize) + ' B';
end;

procedure TProcessThread.UpdateStatus;
begin
  MainForm.StatusEd.Text := FStatus;
end;

end.

